// *******************************************************************
// This file contains copyrighted work from The Foundry,
// Sony Pictures Imageworks and Pixar, is intended for
// Katana and PRMan customers, and is not for distribution
// outside the terms of the corresponding EULA(s).
// *******************************************************************

#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <map>
#include <vector>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>

#if _WIN32
#ifdef GetEnvironmentVariable
#undef GetEnvironmentVariable
#endif
#endif //_WIN32

#include "PRManRender.h"
#include "ProducerPacket.h"

#include <PRManRendererCommon.h>
#include <RfKLog.h>

#include <FnAttribute/FnConstVector.h>
#include <FnRendererInfo/plugin/RenderMethod.h>

#include <FnAsset/FnDefaultAssetPlugin.h>
#include <FnAsset/FnDefaultFileSequencePlugin.h>

#include <FnAttribute/FnGroupBuilder.h>

#include <dirent.h>
#include <regex.h>

#include <dlfcn.h>

namespace pypath = pystring::os::path;

using namespace FnKat;

// ------------------------------------------------------------
//  Static Methods
// ------------------------------------------------------------

// Functions to deal with parsing of RiFilter arguments, to allow quoted
// arguments in in 'single quotes' or "double quotes" to be treated as
// single arguments, and to correctly handle escaped characters.

static char handleEscapedChar(char charIn)
{
    char charOut = charIn;
    switch (charIn)
    {
        case 't': charOut = '\t'; break;
        case 'b': charOut = '\b'; break;
        case 'n': charOut = '\n'; break;
        case 'r': charOut = '\r'; break;
        case 'f': charOut = '\f'; break;
    }
    return charOut;
}

static void splitRiFilterArgs(const std::string& args, 
                              std::vector<std::string>& tokens,
                              PRManPluginState& sharedState)     // for warning message
{
    tokens.clear();
    const int argsLength = args.length();
    bool unterminatedStringArg = false;
    std::string token;

    int j = 0;
    while (j < argsLength)
    {
        // Eat up leading whitespace
        while (j < argsLength && isspace(args[j]))
            j++;

        // End of the string?
        if (j >= argsLength)
            break;

        if ((args[j] == '"') || (args[j] == '\''))
        {
            // Enter quote - look for the corresponding end quote character
            const char quoteChar = args[j++];
            token = "";

            while ((j < argsLength) && (args[j] != quoteChar))
            {
                char nextChar = args[j];

                // Handle escaped characters
                if ((nextChar == '\\') && (j < argsLength - 1))
                    nextChar = handleEscapedChar(args[++j]);

                token.push_back(nextChar);
                j++;
            }

            // Was the string terminated?
            if (j >= argsLength)
            {
                unterminatedStringArg = true;
            }
            else
            {
                tokens.push_back(token);
                // Don't include the end quote character
                j++;
            }
        }
        else
        {
            // Not a quoted argument - look for the next space
            token.clear();

            while ((j < argsLength) && (!isspace(args[j])))
            {
                char nextChar = args[j];

                // Handle escaped characters
                if ((nextChar == '\\') && (j < argsLength - 1))
                    nextChar = handleEscapedChar(args[++j]);

                token.push_back(nextChar);
                j++;
            }

            tokens.push_back(token);
        }

        if (unterminatedStringArg)
        {
            std::ostringstream warningStream;
            warningStream << "Invalid RiFilter argument string : " 
                          << args << std::endl;
            sharedState.Report_Warning(warningStream.str(), "");
            break;
        }
    }
}


static void checkForRedirect()
{
    // If this environment variable has been set, redirect streams
    const char* redirect = ::getenv("RFK_REDIRECT_OUTPUT");
    if (! redirect) return;

    // Debug Workaround - redirect std streams from Render log back to
    // the console

    const char* stdo = ::getenv("KATANA_ORIG_STDOUT");
    const char* stde = ::getenv("KATANA_ORIG_STDERR");

    std::cerr << "stdout: " << stdo << std::endl;
    std::cerr << "stderr: " << stde << std::endl;

    std::cout << "Restoring console STDOUT/STDERR. Check console for Render Log..." << std::endl;

    dup2(atoi(stdo), STDOUT_FILENO);
    dup2(atoi(stde), STDERR_FILENO);

    std::cout << "Redirected Render stdout." << std::endl;
    std::cerr << "Redirected Render stderr." << std::endl;

    // End Workaround.
}


// ------------------------------------------------------------
//  Constructor
// ------------------------------------------------------------

PRManRender::PRManRender(FnScenegraphIterator rootIterator, 
                          FnAttribute::GroupAttribute arguments) :
    RenderBase(rootIterator, arguments),
    m_recoverRequested(false),
    m_checkpoint(false),
    m_bake(false),
    m_ribDump(false),
    m_ribDumpExpandProcedural(false)
{
    // Need to know if checkpoint or recover is enabled before
    // configureDiskRenderOutputProcess is called which happens 
    // before start() is called.
    PRManGlobalSettings globalSettings(rootIterator);
    getCheckpointAndRecoverSettings(globalSettings);
    
    // Find out if the hider is bake to avoid copying the non-existent
    // output render 
    StringAttribute hiderAttr = globalSettings.getAttribute("hider.hider");
    if (hiderAttr.getValue("", false) == "bake")
    {
        m_bake = true;
    }

    // Initialize logging system
    RfKLogInit();
}


// ------------------------------------------------------------
//  Virtual Methods
// ------------------------------------------------------------

int PRManRender::start()
{
    // Debugging mode - dump all output to console
    checkForRedirect();

    RfKLogTrace << "[RfKProfile]: PRManRender::start.";

    gettimeofday(&m_startTime, NULL);

    FnScenegraphIterator rootIterator = getRootIterator();

    PRManRenderSettings renderSettings(rootIterator);
    PRManGlobalSettings globalSettings(rootIterator);

    // Secret attribute: sleep for a bit to allow a debugger or profiler to attach
    IntAttribute sleepAttr = globalSettings.getAttribute("sleepAtStartup.secs");
    int ztime = sleepAttr.getValue(0, false);
    if (ztime > 0)
    {
        std::cerr << "\nFYI:  Sleeping at startup for " << ztime << "seconds...\n";
        sleep(ztime);
        std::cerr << "...Awake\n";
    }

    buildSearchPaths();

    // How are we rendering? (need this before setting up display driver)
    m_renderMethodName = getRenderMethodName();

    // Read globalSettings and renderMethodName to derive display driver name
    setDisplayDriverName(globalSettings);

    // Treat batch renders as disk renders
    if (m_renderMethodName == RendererInfo::DiskRenderMethod::kBatchName)
    {
        m_renderMethodName = RendererInfo::DiskRenderMethod::kDefaultName;
    }

    // Initialize procedural parameters from render settings before code 
    // diverges into separate render methods.
    initProceduralSettings(rootIterator, renderSettings);

    // Render!
    if (m_renderMethodName == RendererInfo::LiveRenderMethod::kDefaultName)
    {
        m_sharedState.SetRerenderingActive(true);
        startRerender(rootIterator, renderSettings, globalSettings);
    }
    else
    {
        std::string ribOut = getRenderOutputFile();

        m_sharedState.SetRerenderingActive(false);

        // RiBegin() and RiFrameBegin() called by setupScene
        setupScene(rootIterator, renderSettings, globalSettings, ribOut);

        renderScene(rootIterator, renderSettings);

        m_sharedState.GetRi()->FrameEnd();
    }

    return 0;
}


int PRManRender::stop()
{
    RfKLogTrace << "[RfKProfile]: PRManRender::stop.";

    struct timeval endTime;
    gettimeofday(&endTime, NULL);
    double delta = (endTime.tv_sec - m_startTime.tv_sec)
                 + ((endTime.tv_usec - m_startTime.tv_usec) / 1000000.0f);
    RfKLogInfo << "[RfKProfile]: Render finished in " << delta << " s.";

    // Close the Ri context
    m_sharedState.GetRi()->End();

    // The rib dump is the only render process that is in-process
    // (PRManBegin/PRManEnd) so ending it is omitted if that is
    // the case.
    if (!m_ribDump)
        return m_sharedState.GetRic()->PRManEnd();

    return 0;
}


int PRManRender::stopLiveEditing()
{
    // std::cerr << "\n --- --- PRManRender::stopLiveEditing " << std::endl;

    // Known Katana issue: this is never called when live editing terminates

    m_sharedState.GetRi()->EditWorldEnd();
    return 0;
}


int PRManRender::queueDataUpdates(FnAttribute::GroupAttribute updateAttribute)
{
    if (m_liveRenderUpdate.get())
        m_liveRenderUpdate->processUpdates(updateAttribute);

    return 0;
}


bool PRManRender::hasPendingDataUpdates() const
{
    if (m_liveRenderUpdate.get())
    {
        if (m_liveRenderUpdate->getNumberOfUpdates() > 0)
            return true;
    }

    return false;
}


int PRManRender::applyPendingDataUpdates()
{
    if (m_liveRenderUpdate.get())
        m_liveRenderUpdate->applyUpdates(getRootIterator());

    return 0;
}


void PRManRender::configureDiskRenderOutputProcess(
        FnKatRender::DiskRenderOutputProcess& diskRenderOutputProcess,
        const std::string& outputName,
        const std::string& outputPath,
        const std::string& renderMethodName,
        const float& frameTime) const
{
    // Get the recipe and retrieve the render settings
    FnScenegraphIterator rootIterator = getRootIterator();
    PRManRenderSettings renderSettings(rootIterator);

    // Build temporary render and final target locations
    std::string tempRenderLocation = 
            RenderOutputUtils::buildTempRenderLocation(rootIterator, outputName, 
                                                       "render", "exr", frameTime);
    std::string targetRenderLocation = outputPath;

    if (renderSettings.isTileRender())
        targetRenderLocation = 
                RenderOutputUtils::buildTileLocation(rootIterator, outputPath);

    // Get the attributes for the render output from the recipe
    FnKatRender::RenderSettings::RenderOutput output = 
                renderSettings.getRenderOutputByName(outputName);

    // The render action used for this render output
    std::auto_ptr<FnKatRender::RenderAction> renderAction;

    // Determine the rendering behaviour based on the hider and output type
    if (m_bake)
    {
        renderAction.reset(
            new FnKatRender::TemporaryRenderAction(tempRenderLocation));
        renderAction->setLoadOutputInMonitor(false);
    }
    else if (output.type == "color")
    {
        if (m_checkpoint || m_recoverRequested)
        {
            RfKLogWarning << "Checkpoints and recover will not work with color output.";
        }

        if (renderSettings.isTileRender())
        {
            renderAction.reset(
                new FnKatRender::CopyRenderAction(targetRenderLocation, 
                                                  tempRenderLocation));
        }
        else
        {
            renderAction.reset(
                new FnKatRender::CopyAndConvertRenderAction(targetRenderLocation,
                                                            tempRenderLocation, 
                                                            output.clampOutput, 
                                                            output.colorConvert, 
                                                            output.computeStats,
                                                            output.convertSettings));
        }
    }
    else if (output.type == "deep")
    {
        renderAction.reset(
                new FnKatRender::CopyRenderAction(targetRenderLocation, 
                                                  tempRenderLocation));
        renderAction->setLoadOutputInMonitor(false);
    }
    else if (output.type == "shadow")
    {
        if (renderSettings.isTileRender())
        {
            renderAction.reset(
                new FnKatRender::CopyRenderAction(targetRenderLocation, 
                                                  tempRenderLocation));
        }
        else
        {
            std::string shadowType = 
                    getStringAttrValue(output.rendererSettings["shadowType"]);
            if (shadowType == "zfile")
            {
                renderAction.reset(
                    new FnKatRender::TemporaryRenderAction(tempRenderLocation));

                if (output.fileExtension == "tx")
                    diskRenderOutputProcess.addPostCommand("txmake -shadow " + tempRenderLocation + 
                                                            std::string(" ") + targetRenderLocation);
                else if (output.fileExtension == "exr")
                    diskRenderOutputProcess.addPostCommand("txmake -shadow -format openexr " + tempRenderLocation + 
                                                            std::string(" ") + targetRenderLocation);
                else if (output.fileExtension == "tif" || output.fileExtension == "tiff")
                    diskRenderOutputProcess.addPostCommand("txmake -shadow -format tiff " + tempRenderLocation + 
                                                            std::string(" ") + targetRenderLocation);
            }
            else if (shadowType == "deepshad")
            {
                renderAction.reset(
                    new FnKatRender::CopyRenderAction(targetRenderLocation, 
                                                      tempRenderLocation));
            }
        }

        renderAction->setLoadOutputInMonitor(false);
    }
    else if (output.type == "raw")
    {
        int rawHasOutput = getAttrValue<int, FnAttribute::IntAttribute>
                                (output.rendererSettings["rawHasOutput"], 0);
        if (rawHasOutput)
        {
            if (m_checkpoint || m_recoverRequested)
            {
                // If checkpoint or recovery is enabled then render directly to
                // the target render location instead of letting Katana render
                // to the tmp file then copying.  We need to keep the original
                // rendered file so prman can pick up where it left off.
                renderAction.reset(
                    new FnKatRender::RenderAction(targetRenderLocation));
            }
            else
            {
                renderAction.reset(
                    new FnKatRender::CopyRenderAction(targetRenderLocation, 
                                                      tempRenderLocation));
            }
        }
        else
        {
            renderAction.reset(
                new FnKatRender::TemporaryRenderAction(tempRenderLocation));
            renderAction->setLoadOutputInMonitor(false);
        }
    }
    else if (output.type == "ptc")
    {
        tempRenderLocation = 
                DefaultAssetPlugin::resolvePath(output.tempRenderLocation, 
                                                (int) getRenderTime());

        bool inValidRange = false;
        std::string validFrames = 
                getStringAttrValue(output.rendererSettings["validFrames"]);
        if (validFrames == "All")
        {
            inValidRange = true;
        }
        else if (validFrames == "None")
        {
            inValidRange = false;
        }
        else
        {
            std::string validRange = 
                getStringAttrValue(output.rendererSettings["validFrameRange"]);
            std::string seqStr = 
                std::string("image.(") + validRange + std::string(")%04d.exr");
            inValidRange = 
                DefaultFileSequencePlugin::isFrameInFileSequence(seqStr, (int) frameTime);
        }

        if (inValidRange)
        {
            renderAction.reset(
                new FnKatRender::CopyRenderAction(targetRenderLocation, 
                                                  tempRenderLocation));
        }
        else
        {
            renderAction.reset(new FnKatRender::NoOutputRenderAction());
            std::cout << "Warning: Not copying '" << tempRenderLocation 
                      << "' as it is not in the valid range." << std::endl;
        }

        renderAction->setLoadOutputInMonitor(false);
    }
    else
    {
        renderAction.reset(new FnKatRender::NoOutputRenderAction());
        std::cout << "Warning: Unknown render output type '" << output.type 
                << "' for render output '" << outputName << "'." << std::endl;
    }

    diskRenderOutputProcess.setRenderAction(renderAction);
}



// ------------------------------------------------------------
//  Render startup
// ------------------------------------------------------------


int PRManRender::setupScene(FnScenegraphIterator rootIterator, 
                             PRManRenderSettings& renderSettings,
                             PRManGlobalSettings& globalSettings, 
                             std::string ribFile)
{
    Ri* ri = m_sharedState.GetRi();
    if (ribFile != "")
    {
        m_ribDump = true;
        m_ribDumpExpandProcedural = isExpandProceduralActive();
        m_sharedState.Debug("PRManRender::setupScene calling RiBegin(\"%s\")", 
                            ribFile.c_str());
        ri->Begin((char*) ribFile.c_str());
    }
    else
    {
        // PRManBegin, RiBegin
        m_sharedState.Debug("PRManRender::setupScene calling beginDirectRender");
        beginDirectRender(rootIterator, renderSettings, globalSettings, false);

        // Finish initialization now that we have a rendering context
        m_sharedState.InitRenderManPostBegin();
    }

#ifndef _WIN32
    ri->System((char*)"/bin/echo Render process started.");
#endif

    // Set error handling before any other PRMan command
    addErrorHandler(globalSettings);

    addCheckpointOptions(globalSettings);

    addFrameCamera(rootIterator, globalSettings, renderSettings);

    // RiFrameBegin
    int frameNumber = (int) getRenderTime();
    ri->FrameBegin((RtInt) frameNumber);

    addRibInclude(rootIterator, globalSettings, "After FrameBegin; Before Display");

    // XXX if (renderSettings.isValid())

    addShutter(renderSettings);
    addSearchPaths(globalSettings);
    globalSettings.apply(&m_sharedState);
    
    // Add the rest of the global Ri state
    addFormat(renderSettings);
    addAdditionalCameras(renderSettings);
    addDisplayLines(renderSettings);
    addQuantize(globalSettings);

    addRibInclude(rootIterator, globalSettings, "After Display; Before Camera");

    // Get camera fov motion blur flag
    GroupAttribute globalSettingsAttr = globalSettings.getGlobalSettingsAttr();
    GroupAttribute cameraAttr = globalSettingsAttr.getChildByName("camera");
    GroupAttribute motionBlurAttr = cameraAttr.getChildByName("motionBlur");
    IntAttribute motionBlurFovAttr = motionBlurAttr.getChildByName("fov");

    bool fovMotionBlur = false;
    if(motionBlurFovAttr.isValid()) 
    {
        fovMotionBlur = motionBlurFovAttr.getValue(false, false);
    }

    // add camera projection 
    int useDefaultProjection = 
            addCameraProjection(renderSettings.getCameraSettings(), 
                                globalSettings);

    addCamera(renderSettings.getCameraSettings(), 
              renderSettings, true, useDefaultProjection, fovMotionBlur);
    
    addRibInclude(rootIterator, globalSettings, "After Camera; Before WorldBegin");

    addClippingPlanes(rootIterator, renderSettings);

    return 0;
}


int PRManRender::setupRerender(FnScenegraphIterator rootIterator, 
                                PRManRenderSettings& renderSettings, 
                                PRManGlobalSettings& globalSettings)
{
    // PRManBegin, RiBegin
    beginDirectRender(rootIterator, renderSettings, globalSettings, true);

    // Options, etc
    addSearchPaths(globalSettings);
    globalSettings.apply(&m_sharedState);

    // RiDisplay, RiCropWindow
    addDisplayLines(renderSettings);
    addCropWindow(renderSettings);

    addLodrange(globalSettings);

    return 0;
}



void PRManRender::fillCommandLineArguments(
            std::vector<std::string>& commandLineArgs, 
            FnScenegraphIterator sgi, 
            PRManRenderSettings& renderSettings,
            PRManGlobalSettings& globalSettings, 
            bool fromLiveRender)
{
    // PRManBegin() looks at argv[0] to determine how it should be initialized.
    // If it ends in 'prman' then it will adjust process resource limits, and
    // install prman's custom signal handlers.  Otherwise it assumes the
    // calling application will take care of all of that.  Since we want
    // 'renderboot' to act like 'prman' in this respect, we're passing 'prman'
    // as argv[0].
    // commandLineArgs.push_back("");
    commandLineArgs.push_back("prman");

    // Get the attribute that contains the prman command line settings
    FnAttribute::GroupAttribute commandArgsSettings = 
                    globalSettings.getAttribute("commandArguments");

    // If "options.limits.threads" setting was overridden add "-t:#"
    int numberOfRenderThreads=0;
    if (getThreadLimitOverride(renderSettings, globalSettings, 
                               commandArgsSettings, numberOfRenderThreads))
    {
        // Print a message in the render log (used by the test harness)
        char renderThreadsBuffer[1024];
        sprintf(renderThreadsBuffer, "%d", numberOfRenderThreads);
        std::cout << "Using " << renderThreadsBuffer << " threads." << std::endl;

        std::string renderThreadDeclaration = 
                        std::string("-t:") + std::string(renderThreadsBuffer);
        commandLineArgs.push_back(renderThreadDeclaration);
    }

    // RIB capture path
    FnAttribute::StringAttribute capturePathAttr = 
                    commandArgsSettings.getChildByName("capturePath");
    if (capturePathAttr.isValid())
    {
        std::string capturePath = capturePathAttr.getValue();
        commandLineArgs.push_back("-capture");
        commandLineArgs.push_back(capturePath);
    }

    // RI Filters
    StringAttribute riFiltersAttr = 
                    commandArgsSettings.getChildByName("riFilters");
    if (riFiltersAttr.isValid())
    {
        // We can have a number of ri filters defined in a single string, 
        // separated by |
        //   e.g.: 'myFilter1|myFilter2 arg1 arg2'
        // where the first token in each string is the name of the ri filter,
        // and the remaining tokens are the filter arguments.
        std::vector<std::string> riFilters;
        pystring::split(riFiltersAttr.getValue(), riFilters, "|");

        std::vector<std::string>::iterator filtersIter;
        std::vector<std::string>::iterator filterArgsIt;

        for (filtersIter = riFilters.begin(); 
             filtersIter != riFilters.end(); 
             filtersIter++)
        {
            std::vector<std::string> filterTokens;
            splitRiFilterArgs(*filtersIter, filterTokens, m_sharedState);

            if (filterTokens.size() == 0)
                continue;

            commandLineArgs.push_back("-rif");
            commandLineArgs.push_back(filterTokens[0]);

            // Leave loop if the filter has no arguments
            if (filterTokens.size() == 1)
                continue;

            commandLineArgs.push_back("-rifargs");
            for (filterArgsIt = filterTokens.begin() + 1; 
                 filterArgsIt != filterTokens.end(); 
                 filterArgsIt++)
            {
                commandLineArgs.push_back((*filterArgsIt));
            }
            commandLineArgs.push_back("-rifend");
        }
    }

    // Progress messages in the render log
    if (! fromLiveRender)
    {
        IntAttribute progressAttr = commandArgsSettings.getChildByName("progress");

        int defaultVal = true;
        if (progressAttr.getValue(defaultVal, false))
        {
            commandLineArgs.push_back("-Progress");
        }
    }

    // Warning filters
    StringAttribute woffAttr = commandArgsSettings.getChildByName("woff");
    if (woffAttr.isValid())
    {
        std::vector<std::string> woff;
        pystring::split(woffAttr.getValue(), woff, "|");
        std::vector<std::string>::iterator woffIter;

        for (woffIter = woff.begin(); woffIter != woff.end(); ++woffIter)
        {
            std::string woffToken = (*woffIter);

            if (!woffToken.empty())
            {
                commandLineArgs.push_back("-woff");
                commandLineArgs.push_back(woffToken);
            }
        }
    }
    
    // Check for alternate render display options
    if (m_displayDriverName == "it" && 
         m_renderMethodName != RendererInfo::DiskRenderMethod::kDefaultName)
    {
        commandLineArgs.push_back("-dspyserver");
        commandLineArgs.push_back(m_displayDriverName);
        commandLineArgs.push_back("-__dl");  // specify that we are in direct linked mode
    }
    
    // Only allow recover to be enabled if we are not in a live/preview render
    if (m_recoverRequested)
    {
        if (m_renderMethodName != RendererInfo::PreviewRenderMethod::kDefaultName &&
            m_renderMethodName != RendererInfo::LiveRenderMethod::kDefaultName) 
    {
        commandLineArgs.push_back( "-recover" );
        commandLineArgs.push_back( "1" );
    }
        else 
        {
            // (Keep both until we fully convert)
            m_sharedState.Report_Warning("Disabling recover for %s render.", m_renderMethodName);
            RfKLogWarning << "Disabling recover for " << m_renderMethodName << " render.";
        }
    }
}

static char* convert(const std::string& str)
{
    return (char*) str.c_str();
}

//
//  Initialize the PRMan renderer for direct rendering
//
int PRManRender::beginDirectRender(FnScenegraphIterator sgi, 
                              PRManRenderSettings& renderSettings,
                              PRManGlobalSettings& globalSettings, 
                              bool fromLiveRender)
{
    // Extract prman parameters from relevant settings objects
    std::vector<std::string> commandLineArgs;
    fillCommandLineArguments(commandLineArgs, sgi, renderSettings, 
                             globalSettings, fromLiveRender);

    // Convert std::string arg list into format acceptable by PRManBegin
    std::vector<char*> argsv;
    std::transform(commandLineArgs.begin(), commandLineArgs.end(),
                    std::back_inserter(argsv), convert);

    // Display command line arguments in debug mode
    m_sharedState.Debug("Issue PRManBegin");
    for (unsigned i=0;i<commandLineArgs.size();i++)
        m_sharedState.Debug(" %d: %s", i, argsv[i]);

    // PRManBegin: Create and configure the prman renderer
    int status = m_sharedState.GetRic()->PRManBegin(commandLineArgs.size(), 
                                                    &argsv[0]);
    if (status)
    {
        m_sharedState.Error("PRManBegin error %d", status);
        return status;
    }

    // RiBegin: Set up a new rendering context for direct rendering
    m_sharedState.Debug("Ready for direct rendering via RiBegin(RI_RENDER)");
    m_sharedState.GetRi()->Begin(RI_RENDER);

    m_sharedState.Debug("PRManBegun");

    return 0;
}


int PRManRender::startRerender(FnScenegraphIterator rootIterator, 
                                PRManRenderSettings renderSettings, 
                                PRManGlobalSettings globalSettings)
{
    Ri* ri = m_sharedState.GetRi();

    // Re-rendering is processed in two stages. The first stage involves 
    // rendering a first pass which is stored to a RIB file.
    // The second stage launches another render using the RIB as a starting
    // place for further edits.

    //
    // The first re-render stage: bake RIB
    //

    // Emit initial Ri commands to RIB
    // Includes RiBegin and RiFrameBegin, Displays and Camera
    std::string ribFile = getKatanaTempDirectory()+std::string("/rerender.rib");
    setupScene(rootIterator, renderSettings, globalSettings, ribFile);


    // Now bake the rest of the scene to RIB
    renderScene(rootIterator, renderSettings);
    ri->FrameEnd();
    ri->End();      // ri context is now invalid until setupRerender called

    // The rib dump is the only render process that is in-process
    // (PRManBegin/PRManEnd) so ending it is omitted if that is the case.
    if (!m_ribDump)
    {
        m_sharedState.GetRic()->PRManEnd();
    }


    //
    // The second re-render stage: EditWorldBegin
    //

    // Initialize prman, emit global settings incl. Display lines
    setupRerender(rootIterator, renderSettings, globalSettings);

    IntAttribute progressiveAttr =
            globalSettings.getAttribute("rerender.progressive");
    int progressive = progressiveAttr.getValue(1, false);

    // Create PRManLiveRender class instance
    m_liveRenderUpdate.reset(
            new PRManLiveRender(&m_sharedState,
                                globalSettings.getGlobalSettingsAttr()));

    // EditWorldBegin
    {
        // Not seen in rerender.rib journal output, print to console for debug
        char buf[512];
        sprintf(buf, "EditWorldBegin \"%s\" \"%s\" %d\n",
                          ribFile.c_str(),
                          "int progressive", progressive);

        m_sharedState.Info("%s", buf);

        // prevent the render from starting immediately so the we can issue
        // an "initialMute" update command
        RtFloat doRender = 0.0f;

        // Start re-render
        ri->EditWorldBegin((char*) ribFile.c_str(),
                          "int progressive", &progressive,
                          "float dorender", &doRender,
                          RI_NULL);
    }
        
    // Build list of muted light locations 
    std::vector<std::string> mutedLocations;
    m_sharedState.GetMutedLightLocations(mutedLocations);
    
    // Manually push an update edit for muted lights so they start in the
    // correct illuminate override state
    m_liveRenderUpdate->processUpdates(
        FnAttribute::GroupBuilder()
            .set("i0.type", FnAttribute::StringAttribute("initialMute"))
            .set("i0.location", FnAttribute::StringAttribute("/root/world"))
            .set("i0.attributes.lightPaths",
                    FnAttribute::StringAttribute(mutedLocations))
            .build());

    return 0;
}


int PRManRender::renderScene(FnScenegraphIterator sgi, 
                             PRManRenderSettings& renderSettings)
{
    // Need to write out imager shaders before WorldBegin.
    m_sharedState.WriteRI_ImagerShaders(sgi);

    Ri* ri = m_sharedState.GetRi();
    ri->WorldBegin();

    RtBound bounds = { 0.0, 1.0, 0.0, 1.0, 0.0, 1.0 };

    m_sharedState.Debug("PRManRender::renderScene writing lights");

    // Write out all lights before going into main procedural.
    // This allows PRMan to include all lights in data structures needed by
    // instancing and grouping.
    FnScenegraphIterator worldSgIterator = sgi.getByPath("/root/world");
    // Want lights to be able to refer to coord sys by name
    m_sharedState.WriteRI_CoordinateSystems(worldSgIterator);
    m_sharedState.WriteRI_Lights(worldSgIterator);

    m_sharedState.Debug("PRManRender::renderScene ribDump %d, "
                        "ribDumpExpandProcedural %d", 
                        m_ribDump, m_ribDumpExpandProcedural);

    //
    //  Three different ways of generating remaining rib:
    //      * writing rib and expanding procedural (e.g. debug rib dump)
    //      * writing rib including procedural (e.g. live render bake)
    //      * direct render via RiProcedural
    //
    if (m_ribDump)
    {
        std::string proceduralArguments = getProceduralArguments(renderSettings);
        if (m_ribDumpExpandProcedural)
        {
            // RIB dump
            // Write expanded procedural to RIB file as DelayedReadArchive 
            // Call ConvertParameters directly and pass returned data to Subdivide
            m_sharedState.Subdivide(m_sharedState.ConvertParameters(
                            const_cast<char*>(proceduralArguments.c_str())), 0);
        }
        else
        {
            // Live render bake: write RiProcedural "DynamicLoad" to rerender.rib

#ifdef _WIN32
            std::string katanaProcedural = m_katanaPluginDir + "\\KatanaProcedural.dll";
#else
            std::string katanaProcedural = m_katanaPluginDir + "/KatanaProcedural.so";
#endif

            // Request correct "DynamicLoad" enum symbol from libprman.so
            RtProcSubdivFunc procDynamicLoad = (RtProcSubdivFunc)
                    m_sharedState.ResolveRiSymbol(
                            "DynamicLoad", PRManPluginState::k_ProceduralSymbol);
            if (!procDynamicLoad)
               m_sharedState.Error("Can't resolve RiProcDynamicLoad");

            // Top-level procedural is cracked immediately
            RtInt yes = 1;
            ri->Attribute((char*)"procedural", 
                          (char*)"int immediatesubdivide", &yes, RI_NULL);
            RtString args[] = { (char*) katanaProcedural.c_str(), 
                                (char*) proceduralArguments.c_str() };

            // Invoke dynamic procedural 
            // ConvertParameters indirectly called via Subdivide
            ri->Procedural((RtPointer) args, bounds, procDynamicLoad, NULL);
        }
    }
    else
    {
        // Direct render: Goes directly into Subdivide via static call on
        // PRManPluginState.  As a result ConvertParameters is not called
        // so it relies on initProceduralSettings to initialize 
        // procedural parameters.

        // Create the producer packet
        PRManProducerPacket* producerPacket = new PRManProducerPacket;
        producerPacket->sgIterator = sgi;
        producerPacket->fullRecurse = m_sharedState.proceduralSettings.fullRecurse;

        // Force initial procedural to be cracked.
        RtInt yes = 1;
        ri->Attribute((char*)"procedural", 
                      (char*)"int immediatesubdivide", &yes, RI_NULL);
        ri->Procedural(producerPacket, bounds, 
                       PRManPluginState::IndirectSubdivide, 
                       PRManPluginState::IndirectFree);
    }

    m_sharedState.Info("PRManRender::renderScene calling RiWorldEnd");
    ri->WorldEnd();

    return 0;
}


// ------------------------------------------------------------
//  Private state initialization
// ------------------------------------------------------------

int PRManRender::buildSearchPaths()
{
    m_katanaBase      = getKatanaBasePath();        // katanaRoot
    m_katanaPluginDir = getRenderPluginDirectory(); // where plugin resides on disk

    m_rmanDsoPath       = getEnvironmentVariable("RMAN_DSOPATH");
    m_rmanShaderPath    = getEnvironmentVariable("RMAN_SHADERPATH");
    m_rmanRixPluginPath = getEnvironmentVariable("RMAN_RIXPLUGINPATH");

    // Set up default searchpath strings if they are not in the environment
    // TODO this is also required in PRManRendererInfo; create rfk.ini ?
    std::string rmantree     = getEnvironmentVariable("RMANTREE");
    std::string rmantree_lib = pypath::join(rmantree, "lib");
    if (m_rmanDsoPath.empty())
        m_rmanDsoPath        = pypath::join(rmantree_lib, "plugins");
    if (m_rmanShaderPath.empty())    
        m_rmanShaderPath     = pypath::join(rmantree_lib, "shaders");
    if (m_rmanRixPluginPath.empty()) 
        m_rmanRixPluginPath  = pypath::join(rmantree_lib, "plugins");

    std::cerr << "m_rmanDsoPath: " << m_rmanDsoPath << std::endl;
    std::cerr << "m_rmanShaderPath: " << m_rmanShaderPath << std::endl;
    std::cerr << "m_rmanRixPluginPath: " << m_rmanRixPluginPath << std::endl;

    // We assume some modules such as the display drivers are relative to the
    // plug-in path (../RenderBin)
    std::string dsoPath(m_katanaPluginDir);
    if (dsoPath[dsoPath.size()-1] != '/') dsoPath += '/';
    m_katanaPluginBinPath = dsoPath + "../RenderBin";

    // XXX We are adding the default dsopath to counteract the fact that Katnana
    // sets the old default $RMAN_DSOPATH at startup. Remove after Foundry fixes Bug 51777.
    m_rmanDsoPath = m_katanaPluginDir + ":" + 
                    m_katanaPluginBinPath + ":" + 
                    m_rmanDsoPath + ":" +
                    pypath::join(rmantree_lib, "plugins");

    return 0;
}

void PRManRender::setDisplayDriverName(PRManGlobalSettings& globalSettings)
{
    std::string renderDisplay = "Monitor";

    // Read prmanGlobalSettings for chosen render output display
    StringAttribute renderDisplayAttr = 
                    globalSettings.getAttribute("renderDisplay");
    if (renderDisplayAttr.isValid())
    {
        renderDisplay = renderDisplayAttr.getValue(renderDisplay, false);
    }

    // Check for alternate render display options
    if (renderDisplay == "it")
    {
        // this also triggers an extra set of prman command line arguments
        m_displayDriverName = renderDisplay;
    }
    else
    {
        // Whether the result is "Monitor" or something unknown, use Monitor settings
        if (m_renderMethodName == RendererInfo::PreviewRenderMethod::kDefaultName)
            m_displayDriverName = "SocketDisplayDriver";
        else
            m_displayDriverName = "SocketMultiresDisplayDriver";

        // Now warn if something else was specified
        if (renderDisplay != "Monitor")
        {
            m_sharedState.Report_Warning("Invalid display driver specified:  "+
                    renderDisplay+".  Falling back to Monitor", "");
        }
    }
}


void PRManRender::getCheckpointAndRecoverSettings(
                PRManGlobalSettings& globalSettings)
{
    // There are two ways to enable prman recover mode: via
    // PrmanGlobalSettings for interactive disk rendering and via an
    // environment variable for farm rendering
    IntAttribute recoverAttr = 
                    globalSettings.getAttribute("commandArguments.recover");
    std::string recoverEnvValue = getEnvironmentVariable( "RFK_RMAN_RECOVER" );
    if (recoverAttr.getValue(false, false) || !recoverEnvValue.empty())
    {
        // If either are set then we'll try to enable recover once we know
        // whether we're doing live/preview or disk/batch rendering.
        m_recoverRequested = true;
    }


    // Grab attribute parent for checkpointing options
    FnAttribute::GroupAttribute globalSettingsAttr = 
                globalSettings.getGlobalSettingsAttr();
    FnAttribute::GroupAttribute checkpointAttr = 
                globalSettingsAttr.getChildByName("options.checkpoint");
    if (checkpointAttr.isValid())
    {
        // options.checkpoint.interval
        FnAttribute::StringAttribute intervalAttr = 
                    checkpointAttr.getChildByName("interval");
        if (intervalAttr.isValid())
        {
            m_checkpointInterval = intervalAttr.getValue("0", false);
        }

        // options.checkpoint.exitat
        FnAttribute::StringAttribute exitatAttr = 
                    checkpointAttr.getChildByName("exitat");
        if (exitatAttr.isValid())
        {
            // Ri option can be string, float or int.  We configure as a
            // string and pass value straight through for prman to parse
            m_checkpointExitat = exitatAttr.getValue("0", false);
        }
        else
        {
            // support the legacy float implementation
            FnAttribute::FloatAttribute exitatFloatAttr = 
                    checkpointAttr.getChildByName("exitat");
            float val = exitatFloatAttr.getValue(0, false);
            if (val != 0)
            {
                char buf[32];
                sprintf(buf, "%f", val);
                m_checkpointExitat = val;
            }
        }

        float interval = strtod(m_checkpointInterval.c_str(), NULL);
        float exitat = strtod(m_checkpointExitat.c_str(), NULL);
        if (interval != 0 || exitat != 0)
            m_checkpoint = true;
    }
}


bool
PRManRender::getThreadLimitOverride(PRManRenderSettings& renderSettings,
                                    PRManGlobalSettings& globalSettings, 
                                    FnAttribute::GroupAttribute& commandArgsSettings,
                                    int& numThreads)
{
    int numberOfRenderThreads = 0;

    // Check render settings for render threads override
    // Returns true if the attribute "renderSettings.renderThreads" is set
    bool renderThreadsSet = renderSettings.applyRenderThreads(numberOfRenderThreads);

    // Check global settings for a render threads override
    // PrmanGlobalSettings commandArguments.numThreadModeProcs 
    // (override to PGS options.limits.threads)
    bool numThreadModeProcsValid = false;
    if (commandArgsSettings.isValid())
    {
        FnAttribute::IntAttribute threadsGlobalAttr = 
                        commandArgsSettings.getChildByName("numThreadModeProcs");
        if (threadsGlobalAttr.isValid())
        {
            numberOfRenderThreads = threadsGlobalAttr.getValue();
            numThreadModeProcsValid = true;
        }
    }

    // Check if the UI preferences or the batch command line contains a render
    // thread override.
    // Returns true if prefs interactiveRenderThreadsOverride is set
    // or --threads3d was passed on the katana --batch command line
    bool threadsOverrideFound = applyRenderThreadsOverride(numberOfRenderThreads);

    // Return final num threads and if it has been overridden
    numThreads = numberOfRenderThreads;
    return ( renderThreadsSet || numThreadModeProcsValid || threadsOverrideFound );
}


// Convert render settings to string parameter string.  Param string is passed
// through to ConvertParameters (in ribDump mode) then is processed into
// ProceduralSettings data.
std::string PRManRender::getProceduralArguments(PRManRenderSettings& renderSettings)
{
    ProceduralArguments arguments;

    Render::addArgument(arguments, "geolib3OpTree", findArgument("geolib3OpTree"));
    Render::addArgument(arguments, "rmanShaderPath", m_rmanShaderPath);
    Render::addArgument(arguments, "rmanRixPluginPath", m_rmanRixPluginPath);
    Render::addArgument(arguments, "katanaBase", m_katanaBase);

    if (m_ribDump)
    {
        if (getRenderMethodName() != RendererInfo::LiveRenderMethod::kDefaultName)
        {
            Render::addArgument(arguments, "ribDump", "", true);

            if (useRenderPassID())
            {
                Render::addArgument(arguments, "useIDPass", "", true);
            }
        }
    }

    if (m_ribDumpExpandProcedural)
    {
        Render::addArgument(arguments, "fullRecurse", "", true);
    }

    Render::addArgument(arguments, "frame", findArgument("renderTime"));
    Render::addArgument(arguments, "shutter", renderSettings.getShutterAsString());
    Render::addArgument(arguments, "cropwindow", renderSettings.getCropWindowAsString());

    Render::addArgument(arguments, "xres", renderSettings.getDataWindowWidthAsString());
    Render::addArgument(arguments, "yres", renderSettings.getDataWindowHeightAsString());

    // If everything has gone well then serialize the arguments
    std::string proceduralArguments = "";
    Render::serialiseCommandLineArguments(arguments, proceduralArguments);

    return proceduralArguments;
}


// Use render settings to to set up katana procedural parameters.  Called ahead
// of code split for render method which guarantees that ProceduralSettings
// will get initial render settings in all cases.
void PRManRender::initProceduralSettings(FnScenegraphIterator& sgi, 
                                        PRManRenderSettings& renderSettings)
{
    float cropWindow[4];
    renderSettings.getCropWindow(cropWindow);
    int dataWindowSize[2];
    renderSettings.getDataWindowSize(dataWindowSize);

    m_sharedState.InitProceduralSettings(sgi,
                                         getRenderTime(),
                                         renderSettings.getShutterOpen(),
                                         renderSettings.getShutterClose(),
                                         cropWindow, 
                                         dataWindowSize
                                         );
}




// ------------------------------------------------------------
//  Methods to emit world-scope Ri settings
// ------------------------------------------------------------

int PRManRender::addSearchPathOption(const char* paramName,
                                      const std::string& envpath,
                                            std::string& searchpath)
{
    Ri* ri = m_sharedState.GetRi();

    // Expand any environment variables in path
    if (! PRManCommonExpandPath(searchpath))
    {
        m_sharedState.Report_Warning(
                "Invalid environment variable in path:  "+searchpath);
    }

    // build path from potential env var and/or path from global settings
    if (!envpath.empty())
    {
        if (!searchpath.empty()) searchpath += std::string(":") + envpath;
        else                       searchpath  = envpath;
    }

    // if we have a searchpath, send it out
    if (!searchpath.empty())
    {
        RtString val[]   = { (char*) searchpath.c_str() };
        ri->Option((char*)"searchpath", paramName, (RtPointer)val, RI_NULL);
    }

    return 0;
}


int PRManRender::addSearchPaths(PRManGlobalSettings &globalSettings)
{
    // Pull these global settings values and combine with any env vars set.
    std::string displayPath;
    std::string proceduralPath;
    std::string shaderPath;
    std::string rixPluginPath;

    GroupAttribute optionsAttr = globalSettings.getAttribute("options");
    if (optionsAttr.isValid())
    {
        GroupAttribute searchPathAttr = optionsAttr.getChildByName("searchpath");
        if (searchPathAttr.isValid())
        {
            StringAttribute displayPathAttr    = searchPathAttr.getChildByName("display");
            StringAttribute proceduralPathAttr = searchPathAttr.getChildByName("procedural");
            StringAttribute shaderPathAttr     = searchPathAttr.getChildByName("shader");
            StringAttribute rixPluginPathAttr  = searchPathAttr.getChildByName("rixplugin");

            displayPath    = displayPathAttr.getValue("", false);
            proceduralPath = proceduralPathAttr.getValue("", false);
            shaderPath     = shaderPathAttr.getValue("", false);
            rixPluginPath  = rixPluginPathAttr.getValue("", false);
        }
    }

    addSearchPathOption("string display",    m_rmanDsoPath, displayPath);
    addSearchPathOption("string procedural", m_rmanDsoPath, proceduralPath);
    addSearchPathOption("string shader",     m_rmanShaderPath, shaderPath);
    addSearchPathOption("string rixplugin",  m_rmanRixPluginPath, rixPluginPath);

    // if shader paths were combined, save the new path
    m_rmanShaderPath    = shaderPath;
    m_rmanRixPluginPath = rixPluginPath;

    return 0;
}


int PRManRender::addErrorHandler(PRManGlobalSettings& globalSettings)
{
    Ri* ri = m_sharedState.GetRi();

    FnAttribute::StringAttribute errorHandlerAttr = 
                    globalSettings.getAttribute("errorHandler");

    if (errorHandlerAttr.isValid())
    {
        const std::string errorHandler = 
                        errorHandlerAttr.getValue("print", false);

        if (errorHandler == "ignore")
        {
            ri->ErrorHandler(RiErrorIgnore);
        }
        else if (errorHandler == "print")
        {
            ri->ErrorHandler(RiErrorPrint);
        }
        else if (errorHandler == "printonce")
        {
            ri->ErrorHandler(RiErrorPrintOnce);
        }
        else if (errorHandler == "abort")
        {
            ri->ErrorHandler(RiErrorAbort);
        }
        else
        {
            m_sharedState.Report_Warning("Invalid error handler specified.", "");
        }
    }

    return 0;
}

int PRManRender::addRibInclude(FnScenegraphIterator sgi, 
                                PRManGlobalSettings& globalSettings, 
                                std::string currentLocation)
{
    FnAttribute::GroupAttribute globalSettingsAttr = 
                    globalSettings.getGlobalSettingsAttr();

    FnAttribute::StringAttribute ribIncludeAttr = 
                    globalSettingsAttr.getChildByName("ribInclude");
    if (ribIncludeAttr.isValid())
    {
        std::string location;
        StringAttribute ribIncludeLocationAttr = 
                    globalSettingsAttr.getChildByName("ribIncludeLocation");
        if (ribIncludeLocationAttr.isValid())
            location = ribIncludeLocationAttr.getValue();
        else
            location = "After FrameBegin; Before Display";

        if (location != currentLocation)
        {
            return 0;
        }

        std::string data = ribIncludeAttr.getValue();
        m_sharedState.WriteRI_RibInclude(data);
    }
    else
    {
        return 1;
    }

    return 0;
}

int PRManRender::addClippingPlanes(FnScenegraphIterator sgi, 
                                    PRManRenderSettings& renderSettings)
{
    FnScenegraphIterator worldIterator = sgi.getByPath("/root/world");
    if (!worldIterator.isValid())
        return 1;

    Ri* ri = m_sharedState.GetRi();
    GroupAttribute globalsAttr = worldIterator.getAttribute("globals");
    if (globalsAttr.isValid())
    {
        GroupAttribute clippingPlanesAttr = 
                        globalsAttr.getChildByName("clippingPlanes");
        if (clippingPlanesAttr.isValid())
        {
            int noClippingPlanes = clippingPlanesAttr.getNumberOfChildren();
            for (int i = 0; i < noClippingPlanes; ++i)
            {
                StringAttribute planeAttr = 
                                clippingPlanesAttr.getChildByIndex(i);
                if (planeAttr.isValid())
                {
                    std::string clippingPlaneName = planeAttr.getValue();
                    FnScenegraphIterator planeProducer = 
                                    sgi.getByPath(clippingPlaneName);
                    if (!planeProducer.isValid())
                        continue;

                    // TODO convert this into a ClippingPlaneResolver Op to
                    // move this attr info local to the place which needs it...
                    // very similar to the relativeScopedCoordinateSystem thing
                    // in that it needs a ws xform and a little more info (but
                    // not much)
                    const std::string clipLocation = clippingPlaneName;
                    const FnAttribute::GroupAttribute clipLocationAttr =
                                                planeProducer.getAttribute("");

                    IntAttribute enableAttr = clipLocationAttr.getChildByName(
                                                    "enableClippingPlane");
                    if (!enableAttr.isValid())
                        continue;
                    if (enableAttr.getValue() == 0)
                        continue;

                    bool reversePlane = false;
                    IntAttribute reverseAttr = clipLocationAttr.getChildByName(
                                                    "reverseClippingDirection");
                    if (reverseAttr.isValid() && reverseAttr.getValue() == 1)
                        reversePlane = true;

                    ri->TransformBegin();
                    m_sharedState.WriteRI_Object_GlobalTransform(
                                                clipLocation, clipLocationAttr);

                    if (reversePlane)
                        ri->ClippingPlane(0, -1, 0, 0, 0, 0);
                    else
                        ri->ClippingPlane(0, 1, 0, 0, 0, 0);

                    ri->TransformEnd();
                }
            }
        }
    }

    return 0;
}

int PRManRender::addShutter(PRManRenderSettings& renderSettings)
{
    if (renderSettings.getMaxTimeSamples() > 1)
    {
        m_sharedState.GetRi()->Shutter(renderSettings.getShutterOpen(), 
                                       renderSettings.getShutterClose());
    }

    return 0;
}

int PRManRender::addFormat(PRManRenderSettings& renderSettings)
{
    int dataWindowSize[2];
    renderSettings.getDataWindowSize(dataWindowSize);
    float pixelaspectratio = 1.f;

    Ri* ri = m_sharedState.GetRi();
    ri->Identity();
    ri->Orientation(RI_RH);
    ri->Format(dataWindowSize[0], dataWindowSize[1], pixelaspectratio);

    return 0;
}

void appendChildAttrsToAttrList(PRManPluginState& sharedState, 
                                 PRManAttrList& attrList, 
                                 FnAttribute::GroupAttribute groupAttr,
                                 const std::string& namePrefix = "")
{
    for (int i = 0; i < groupAttr.getNumberOfChildren(); ++i)
    {
        const std::string childName = namePrefix + groupAttr.getChildName(i);
        std::string declaration = sharedState.BuildDeclarationFromAttribute(
                                        childName,
                                        groupAttr.getChildByIndex(i));
        if (!declaration.empty())
        {
            attrList.push_back(PRManAttrListEntry(declaration,
                                        childName,
                                        groupAttr.getChildByIndex(i)));
        }
    }
}

int PRManRender::addDisplayLines(PRManRenderSettings& renderSettings)
{
    if (m_renderMethodName == RendererInfo::PreviewRenderMethod::kDefaultName ||
        m_renderMethodName == RendererInfo::LiveRenderMethod::kDefaultName)
    {
        // Interactive case
        FnKatRender::RenderSettings::ChannelBuffers interactiveBuffers;
        renderSettings.getChannelBuffers(interactiveBuffers);

        // Render outputs
        FnKatRender::RenderSettings::RenderOutputs outputs = 
                        renderSettings.getRenderOutputs();

        std::string underscores = "";
        float sampleRate[2];

        float frameTime = getRenderTime();
        renderSettings.getSampleRate(sampleRate);
        std::string host = getKatanaHost();
        
        int displayWindow[4];
        renderSettings.getDisplayWindow(displayWindow);

        int dataWindow[4];
        renderSettings.getDataWindow(dataWindow);
        
        int allowIDPass = 0;
        if (useRenderPassID())
            allowIDPass = 1;

        int mainSequenceId = -1;
        Ri* ri = m_sharedState.GetRi();

        FnKatRender::RenderSettings::ChannelBuffers::const_iterator it;
        for (it = interactiveBuffers.begin(); 
             it != interactiveBuffers.end(); 
             ++it)
        {
            std::string outputName = it->first;
            
            // XXX, we need to access renderSettings that will be present
            // in the outputs map under the original name. The 0-prefix is a
            // hack to ensure that primary is always first. Previously, this
            // code only needed to grab the camera (which isn't meaningful to
            // the primary output). Because we now need additional information
            // for the interactiveDisplayOptions, let's unhack this here.
            if (outputName == "0primary")
            {
                outputName = "primary";
            }
            
            FnKatRender::RenderSettings::ChannelBuffer buffer = it->second;

            int frameId = atoi(buffer.bufferId.c_str());
            RtString channelName = (char*) buffer.channelName.c_str();
            std::string displayName = underscores + host;

            // Check if there is an alternative camera in the outputs
            std::string cameraName;
            FnKatRender::RenderSettings::RenderOutputs::iterator 
                            outputIt = outputs.find(outputName);
            if (outputIt != outputs.end())
            {
                cameraName = outputIt->second.cameraName;
            }
            
            PRManAttrList attrList;
            
            FnAttribute::GroupAttribute interactiveDisplayOptions;
            
            //check for interactive display options
            if ( outputIt != outputs.end() )
            {
                FnKatRender::RenderSettings::RenderOutput & renderOutput =
                        outputIt->second;
                
                interactiveDisplayOptions =
                        renderOutput.rendererSettings["interactiveDisplayOptions"];
            }
            
            
            attrList.push_back(PRManAttrListEntry("int frameID",
                        "frameID", FnAttribute::IntAttribute(frameId)));
                
            attrList.push_back(PRManAttrListEntry("float frameTime",
                    "frameTime", FnAttribute::FloatAttribute(frameTime)));
            
            attrList.push_back(PRManAttrListEntry("string frameName",
                    "frameName", FnAttribute::StringAttribute(outputName)));
            
            attrList.push_back(PRManAttrListEntry("float[2] sampleRate",
                   "sampleRate", FnAttribute::FloatAttribute(sampleRate, 2, 2)));
            
            attrList.push_back(PRManAttrListEntry("int allowIDPass",
                    "allowIDPass", FnAttribute::IntAttribute(allowIDPass)));
            
            if (!cameraName.empty())
            {
                attrList.push_back(PRManAttrListEntry("string camera",
                        "camera", FnAttribute::StringAttribute(cameraName)));
            }
            
            attrList.push_back(PRManAttrListEntry("int[4] dataWindow",
                   "dataWindow", FnAttribute::IntAttribute(dataWindow, 4, 4)));

            attrList.push_back(PRManAttrListEntry("int[4] displayWindow",
                   "displayWindow", FnAttribute::IntAttribute(displayWindow, 4, 4)));

            appendChildAttrsToAttrList( m_sharedState, attrList, interactiveDisplayOptions );
            
            PRManAttrList_Converter converter(attrList);
            ri->DisplayV( (char*) displayName.c_str(), (char*) m_displayDriverName.c_str(), channelName,
                    converter.getSize(0), converter.getDeclarationTokens(0), converter.getParameters(0) );
            
            
            if (underscores == "")
                underscores = "+";
            underscores += "_";

            if (mainSequenceId == -1)
                mainSequenceId = frameId;
        }

        // Declare an ID pass used for object picking in the Monitor
        if (allowIDPass && mainSequenceId != -1)
        {
            std::string displayName = underscores + host;
            RtString frameName[] = { (char*)"__id" };
            RtString channelName = (char*)"uniform float id";

            int quantize[4] = { 0, 0, 0, 0 };

            RtString filter[] = { (char*)"zmin" };
            float filterWidth[2] = { 1, 1 };

                ri->Display((char*) displayName.c_str(),
                           (char*) m_displayDriverName.c_str(), 
                           channelName,
                           (char*)"int frameID", &mainSequenceId,
                           (char*)"float frameTime", &frameTime,
                           (char*)"string frameName", (RtPointer) frameName,
                           (char*)"quantize", quantize,
                           (char*)"filter", (RtPointer) filter,
                           (char*)"filterwidth", filterWidth,
                           (char*)"float[2] sampleRate", sampleRate,
                           (char*)"int allowIDPass", &allowIDPass,
                           RI_NULL);

            RtString pythonHostName[] = { (char*) host.c_str() };
            ri->Attribute((char*)"user", 
                          (char*)"uniform string ProceduralHostName", 
                          (RtPointer) pythonHostName, RI_NULL);
            ri->Attribute((char*)"user", 
                          (char*)"uniform int ProceduralFrameID", 
                          &mainSequenceId, RI_NULL);
            ri->Attribute((char*)"user", 
                          (char*)"uniform int ProceduralGenerateID", 
                          &allowIDPass, RI_NULL);
        }
    }
    else if (m_renderMethodName == RendererInfo::DiskRenderMethod::kDefaultName)
    {
        FnKatRender::RenderSettings::RenderOutputs outputs = 
                                        renderSettings.getRenderOutputs();
        std::vector<std::string> renderOutputNames = 
                                        renderSettings.getRenderOutputNames();
        std::string displayPrefix = "";
        Ri* ri = m_sharedState.GetRi();

        std::vector<std::string>::const_iterator it;
        for (it = renderOutputNames.begin(); 
             it != renderOutputNames.end(); 
             ++it)
        {
            FnKatRender::RenderSettings::RenderOutput output = outputs[(*it)];

            PRManAttrList attrList;
            std::string displayDriver;
            std::string displayMode;

            // add '+' for secondary
            std::string displayName = displayPrefix + output.renderLocation;    

            FnKatRender::RenderSettings::AttributeSettings settings = 
                                                    output.rendererSettings;

            if (output.type == "color")
            {
                displayDriver = "openexr";
                displayMode = output.channel;
                attrList.push_back(
                    PRManAttrListEntry("string exrcompression",
                                       "exrcompression", 
                                       FnAttribute::StringAttribute("rle")));
            }
            else if (output.type == "deep")
            {
                displayDriver = (output.fileExtension == "deepexr") ? "deepexr" : "deepshad";
                displayMode = output.channel;

                // Check for validity of subimage before adding attr.
                // Passing this as an empty string will overwrite the
                // correct default inside of prman
                std::string subimage = getStringAttrValue( settings["subimage"] );
                if ( !subimage.empty() )
                {
                    attrList.push_back(PRManAttrListEntry("string subimage",
                                       "subimage", settings["subimage"]));
                }
            }
            else if (output.type == "shadow")
            {
                std::string shadowType = getStringAttrValue(settings["shadowType"]);
                if (shadowType == "zfile")
                {
                    displayDriver = "zfile";
                    displayMode = "z";
                }
                else if (shadowType == "deepshad")
                {
                    displayDriver = "deepshad";
                    std::string deepshadMode = 
                            getStringAttrValue(settings["deepshadMode"], 
                                               "deepopacity");
                    displayMode = deepshadMode;
                    attrList.push_back(
                            PRManAttrListEntry("string volumeinterpretation",
                                               "volumeinterpretation", 
                                               settings["deepshadVolumeInterpretation"]));
                }
            }
            else if (output.type == "raw")
            {
                displayDriver = getStringAttrValue(settings["displayDriver"], 
                                                   "openexr");
                displayMode = output.channel;
            }
            else if (output.type == "ptc")
            {
                continue;
            }

            if (displayPrefix != "" && displayMode == "rgba")
            {
                displayMode = "Ci";    // An additional display cannot be rgba
            }

            // Always add an additional camera if it has been declared, regardless of type.
            if (output.cameraName != "")
            {
                attrList.push_back(
                        PRManAttrListEntry("string camera",
                                           "camera", 
                                           FnAttribute::StringAttribute(output.cameraName)));
            }

            // Add custom options to the display line
            FnAttribute::GroupAttribute displayOptionsAttr = 
                                settings["displayOptions"];
            if (displayOptionsAttr.isValid())
            {
                appendChildAttrsToAttrList(m_sharedState, attrList, 
                                           displayOptionsAttr);
            }

            // Starting with PRMan 17, the driver supports the injection of
            // arbitrary metadata. This is done by passing named parameters with
            // the form "exrheader_$key", where $key will be the name of the
            // attribute as it shows up in the OpenEXR header. For example,
            // Display "myfile.exr" "openexr" "rgba" "string exrheader_author" ["Bob"]
            //
            // Note that even though PRMan 17 currently does not support string
            // arrays or floats arrays beyond a size of 4, we still allow
            // passing this options in here since this is likely to be added.
            FnAttribute::GroupAttribute exrheadersAttr = settings["exrheaders"];
            if (exrheadersAttr.isValid())
            {
                appendChildAttrsToAttrList(m_sharedState, attrList, exrheadersAttr, "exrheader_");
            }

            PRManAttrList_Converter converter(attrList);
            ri->DisplayV((char*) displayName.c_str(), 
                         (char*) displayDriver.c_str(), 
                         (char*) displayMode.c_str(),
                         converter.getSize(0), 
                         converter.getDeclarationTokens(0), 
                         converter.getParameters(0));

            if (displayPrefix == "")
            {
                displayPrefix = "+";
            }
        }
    }

    return 0;
}

void PRManRender::addQuantize(PRManGlobalSettings& globalSettings)
{
    bool rgbaQuantizeUsed = false;
    GroupAttribute globalSettingsAttr = globalSettings.getGlobalSettingsAttr();
    GroupAttribute quantizeAttr = globalSettingsAttr.getChildByName("quantize");

    Ri* ri = m_sharedState.GetRi();

    if (quantizeAttr.isValid())
    {
        FnAttribute::FloatAttribute rgbaAttr = quantizeAttr.getChildByName("rgba");
        if (rgbaAttr.isValid())
        {
            FloatConstVector rgbaData = rgbaAttr.getNearestSample(0.f);
            if (rgbaData.size() == 4)
            {
                ri->Quantize(RI_RGBA, 
                              (RtInt) rgbaData[0], (RtInt) rgbaData[1], 
                              (RtInt) rgbaData[2], (RtFloat) rgbaData[3]);
                rgbaQuantizeUsed = true;
            }
        }

        FnAttribute::FloatAttribute zAttr = quantizeAttr.getChildByName("z");
        if (zAttr.isValid())
        {
            FloatConstVector zData = zAttr.getNearestSample(0.f);
            if (zData.size() == 4)
            {
                ri->Quantize(RI_Z, 
                              (RtInt) zData[0], (RtInt) zData[1], 
                              (RtInt) zData[2], (RtFloat) zData[3]);
            }
        }
    }

    if (!rgbaQuantizeUsed)
        ri->Quantize(RI_RGBA, 0, 0, 0, 0);
}


void PRManRender::addCheckpointOptions(PRManGlobalSettings& globalSettings)
{
    // Checkpointing only makes sense during a non-interactive render
    if (m_renderMethodName != 
                RendererInfo::DiskRenderMethod::kDefaultName)
        return;

    if (!m_checkpoint)
        return;

    Ri* ri = m_sharedState.GetRi();

    // Grab attribute parent for checkpointing options
    FnAttribute::GroupAttribute globalSettingsAttr = 
                globalSettings.getGlobalSettingsAttr();
    FnAttribute::GroupAttribute checkpointAttr = 
                globalSettingsAttr.getChildByName("options.checkpoint");

    if (checkpointAttr.isValid())
    {
        // options.checkpoint.interval
        if (!m_checkpointInterval.empty())
        {
            char* val = (char*) m_checkpointInterval.c_str();
            ri->Option((char*)"checkpoint", (char*)"string interval", 
                        (RtPointer)&val, RI_NULL);
        }

        // options.checkpoint.exitat
        if (!m_checkpointExitat.empty())
        {
            char* val = (char*) m_checkpointExitat.c_str();
            ri->Option((char*)"checkpoint", (char*)"string exitat", 
                        (RtPointer)&val, RI_NULL);
        }

        // options.checkpoint.keepfiles
        FnAttribute::IntAttribute keepfilesAttr = 
                    checkpointAttr.getChildByName("keepfiles");
        if (keepfilesAttr.isValid())
        {
            // Simple boolean
            int keepfiles = keepfilesAttr.getValue(0, false);
            ri->Option((char*)"checkpoint", (char*)"int keepfiles", 
                        (RtPointer)&keepfiles, RI_NULL);
        }
        
        // options.checkpoint.command
        FnAttribute::StringAttribute commandAttr = 
                    checkpointAttr.getChildByName("command");
        if (commandAttr.isValid())
        {
            // System command called after checkpoint.
            const std::string command = commandAttr.getValue("", false);
            char* val = (char*) command.c_str();
            ri->Option((char*)"checkpoint", (char*)"string command", 
                        (RtPointer)&val, RI_NULL);
        }
    }
}


int PRManRender::addCropWindow(PRManRenderSettings& renderSettings)
{
    if (! renderSettings.isValid())
        return 1;

    Ri* ri = m_sharedState.GetRi();

    // getCropWindow does the logic to combine the renderSettings
    // dataWindow and regionOfInterest for final cropWindow values
    float cropWindow[4];
    renderSettings.getCropWindow(cropWindow);

    ri->CropWindow(cropWindow[0], cropWindow[1], cropWindow[2], cropWindow[3]);

    return 0;
}

//
// Write a motion block for the animated camera fov.  Since PRMan does not support RiPerspective 
// or RiProjection inside a motion blur.  As a workaround, we are emitting RiScale x y z.  
// RiScale's x and y are set to the following:
//
//    the ratio of tan(fov of the current time step)/over tan(fov from RiProjection).
//
#define radians(x) ((x) * 0.017453292519943295)

void PRManRender::addFOVMotionBlur(
        const DoubleAttribute& fovAttr, float cameraFov)
{
    Ri *ri = m_sharedState.GetRi();

    int64_t numSampleTimes = fovAttr.getNumberOfTimeSamples();

    if( numSampleTimes > 1 )
    {
        std::vector<float> sampleTimes;
        std::vector<float> fovValues;

        for( int64_t sampleIndex = 0; sampleIndex < numSampleTimes; ++sampleIndex )
        {
            float sample = fovAttr.getSampleTime(sampleIndex);
            sampleTimes.push_back(sample);

            DoubleConstVector fovSample = fovAttr.getNearestSample(sample);
            fovValues.push_back(fovSample[0]);
        }

        if (sampleTimes.size() > 1)
        {
            RtInt noSamples = sampleTimes.size();
            ri->MotionBeginV( noSamples, (RtFloat*) &sampleTimes[0] );

            float cameraScale = 1;
            float cameraFovTan = tan(radians(cameraFov*0.5));
            for (int i = 0; i < noSamples; i++) 
            {
                cameraScale = tan(radians(fovValues[i]*0.5))/cameraFovTan;
                ri->Scale(cameraScale, cameraScale, 1);
            }

            ri->MotionEnd();
       }
   }
}


int PRManRender::addDefaultCameraProjection(
                    const FnKatRender::CameraSettings* cameraSettings)
{
    if (!cameraSettings)
        return 1;

    Ri* ri = m_sharedState.GetRi();

    // Projection from CameraCreate
    float fov = cameraSettings->getFov();
    if (cameraSettings->getProjection() == "perspective") 
        ri->Projection(RI_PERSPECTIVE, "fov", &fov, RI_NULL);
    else 
        ri->Projection(RI_ORTHOGRAPHIC, RI_NULL);

    return 0;
}


bool PRManRender::addCameraProjection(
                const FnKatRender::CameraSettings* cameraSettings,
                const PRManGlobalSettings& globalSettings)
{
    bool useDefaultProjection = true;

    Ri *ri = m_sharedState.GetRi();

    GroupAttribute globalSettingsAttr = globalSettings.getGlobalSettingsAttr();
    GroupAttribute cameraAttr = globalSettingsAttr.getChildByName("camera");
    GroupAttribute projectionAttr = cameraAttr.getChildByName("projection");

    StringAttribute projectionTypeAttr = 
                projectionAttr.getChildByName("projectionType");
    if( projectionTypeAttr.isValid())
    {
        // Emit RiProjection to the RIB for the non-default camera projection below.
        std::string projectionType = projectionTypeAttr.getValue("default", false);
        if (projectionType == "cylinder" || 
            projectionType == "sphere" || 
            projectionType == "torus")
        {
            FnAttribute::FloatAttribute hsweepAttr = 
                                projectionAttr.getChildByName("hsweep");
            float hsweep = hsweepAttr.getValue(360.f, false);

            if (projectionType == "cylinder") 
            {
                ri->Projection((char*)"cylinder", (char*)"float hsweep", 
                                &hsweep, RI_NULL);
            } 
            else 
            {
                FnAttribute::FloatAttribute vsweepAttr =
                                    projectionAttr.getChildByName("vsweep");
                float vsweep = vsweepAttr.getValue(180.f, false);
                if (projectionType == "torus") 
                {
                    FnAttribute::FloatAttribute minorAttr = 
                                    projectionAttr.getChildByName("minor");
                    float minor = minorAttr.getValue(0.25f, false);
                    ri->Projection(const_cast<RtToken>(projectionType.c_str()), 
                                   "float hsweep", &hsweep, 
                                   "float vsweep", &vsweep, 
                                   "float minor", &minor, RI_NULL);
                } 
                else 
                {
                    ri->Projection(const_cast<RtToken>(projectionType.c_str()), 
                                   "float hsweep", &hsweep, 
                                   "float vsweep", &vsweep, RI_NULL);
                }

            }

            useDefaultProjection = false;    
        }
        else if (projectionType == "PxrLightProbe") 
        {
            ri->Projection((char*)"PxrLightProbe", RI_NULL);
            useDefaultProjection = false;    
        }
        else if ( projectionType == "bs_tdeLens")
        {
            IntAttribute doDistortAttr = projectionAttr.getChildByName( "doDistort" );
            int doDistort = doDistortAttr.getValue( 1, false);
            IntAttribute modelAttr = projectionAttr.getChildByName( "model" );
            int model = modelAttr.getValue( 1, false);
            StringAttribute distFileAttr = projectionAttr.getChildByName( "distFile" );
            std::string distFile = distFileAttr.getValue( "", false);
            FloatAttribute focal_lengthAttr = projectionAttr.getChildByName( "focal_length" );
            float focal_length = focal_lengthAttr.getValue( 0.0f, false);
            FloatAttribute focus_distanceAttr = projectionAttr.getChildByName( "focus_distance" );
            float focus_distance = focus_distanceAttr.getValue( 0.0f, false);
            FloatAttribute filmBack_WAttr = projectionAttr.getChildByName( "filmBack_W" );
            float filmBack_W = filmBack_WAttr.getValue( 0.0f, false);
            FloatAttribute filmBack_HAttr = projectionAttr.getChildByName( "filmBack_H" );
            float filmBack_H = filmBack_HAttr.getValue( 0.0f, false);
            FloatAttribute xOffsetAttr = projectionAttr.getChildByName( "xOffset" );
            float xOffset = xOffsetAttr.getValue( 0.0f, false);
            FloatAttribute yOffsetAttr = projectionAttr.getChildByName( "yOffset" );
            float yOffset = yOffsetAttr.getValue( 0.0f, false);
            FloatAttribute distortionAttr = projectionAttr.getChildByName( "distortion" );
            float distortion = distortionAttr.getValue( 0.0f, false);
            FloatAttribute anamorphicSqueezeAttr = projectionAttr.getChildByName( "anamorphicSqueeze" );
            float anamorphicSqueeze = anamorphicSqueezeAttr.getValue( 0.0f, false);
            FloatAttribute xCurvatureAttr = projectionAttr.getChildByName( "xCurvature" );
            float xCurvature = xCurvatureAttr.getValue( 0.0f, false);
            FloatAttribute yCurvatureAttr = projectionAttr.getChildByName( "yCurvature" );
            float yCurvature = yCurvatureAttr.getValue( 0.0f, false);
            FloatAttribute quarticDistAttr = projectionAttr.getChildByName( "quarticDist" );
            float quarticDist = quarticDistAttr.getValue( 0.0f, false);
            FloatAttribute dd2Attr = projectionAttr.getChildByName( "dd2" );
            float dd2 = dd2Attr.getValue( 0.0f, false);
            FloatAttribute ud2Attr = projectionAttr.getChildByName( "ud2" );
            float ud2 = ud2Attr.getValue( 0.0f, false);
            FloatAttribute vd2Attr = projectionAttr.getChildByName( "vd2" );
            float vd2 = vd2Attr.getValue( 0.0f, false);
            FloatAttribute qd4Attr = projectionAttr.getChildByName( "qd4" );
            float qd4 = qd4Attr.getValue( 0.0f, false);
            FloatAttribute ud4Attr = projectionAttr.getChildByName( "ud4" );
            float ud4 = ud4Attr.getValue( 0.0f, false);
            FloatAttribute vd4Attr = projectionAttr.getChildByName( "vd4" );
            float vd4 = vd4Attr.getValue( 0.0f, false);
            FloatAttribute d6Attr = projectionAttr.getChildByName( "d6" );
            float d6 = d6Attr.getValue( 0.0f, false);
            FloatAttribute d8Attr = projectionAttr.getChildByName( "d8" );
            float d8 = d8Attr.getValue( 0.0f, false);
            FloatAttribute phiAttr = projectionAttr.getChildByName( "phi" );
            float phi = phiAttr.getValue( 0.0f, false);
            FloatAttribute BAttr = projectionAttr.getChildByName( "B" );
            float B = BAttr.getValue( 0.0f, false);
            FloatAttribute cx02d2Attr = projectionAttr.getChildByName( "cx02d2" );
            float cx02d2 = cx02d2Attr.getValue( 0.0f, false);
            FloatAttribute cy02d2Attr = projectionAttr.getChildByName( "cy02d2" );
            float cy02d2 = cy02d2Attr.getValue( 0.0f, false);
            FloatAttribute cx22d2Attr = projectionAttr.getChildByName( "cx22d2" );
            float cx22d2 = cx22d2Attr.getValue( 0.0f, false);
            FloatAttribute cy22d2Attr = projectionAttr.getChildByName( "cy22d2" );
            float cy22d2 = cy22d2Attr.getValue( 0.0f, false);
            FloatAttribute cx04d4Attr = projectionAttr.getChildByName( "cx04d4" );
            float cx04d4 = cx04d4Attr.getValue( 0.0f, false);
            FloatAttribute cy04d4Attr = projectionAttr.getChildByName( "cy04d4" );
            float cy04d4 = cy04d4Attr.getValue( 0.0f, false);
            FloatAttribute cx24d4Attr = projectionAttr.getChildByName( "cx24d4" );
            float cx24d4 = cx24d4Attr.getValue( 0.0f, false);
            FloatAttribute cy24d4Attr = projectionAttr.getChildByName( "cy24d4" );
            float cy24d4 = cy24d4Attr.getValue( 0.0f, false);
            FloatAttribute cx44d4Attr = projectionAttr.getChildByName( "cx44d4" );
            float cx44d4 = cx44d4Attr.getValue( 0.0f, false);
            FloatAttribute cy44d4Attr = projectionAttr.getChildByName( "cy44d4" );
            float cy44d4 = cy44d4Attr.getValue( 0.0f, false);
            FloatAttribute rotAttr = projectionAttr.getChildByName( "rot" );
            float rot = rotAttr.getValue( 0.0f, false);
            FloatAttribute sqxAttr = projectionAttr.getChildByName( "sqx" );
            float sqx = sqxAttr.getValue( 0.0f, false);
            FloatAttribute sqyAttr = projectionAttr.getChildByName( "sqy" );
            float sqy = sqyAttr.getValue( 0.0f, false);
            FloatAttribute cx06d6Attr = projectionAttr.getChildByName( "cx06d6" );
            float cx06d6 = cx06d6Attr.getValue( 0.0f, false);
            FloatAttribute cy06d6Attr = projectionAttr.getChildByName( "cy06d6" );
            float cy06d6 = cy06d6Attr.getValue( 0.0f, false);
            FloatAttribute cx26d6Attr = projectionAttr.getChildByName( "cx26d6" );
            float cx26d6 = cx26d6Attr.getValue( 0.0f, false);
            FloatAttribute cy26d6Attr = projectionAttr.getChildByName( "cy26d6" );
            float cy26d6 = cy26d6Attr.getValue( 0.0f, false);
            FloatAttribute cx46d6Attr = projectionAttr.getChildByName( "cx46d6" );
            float cx46d6 = cx46d6Attr.getValue( 0.0f, false);
            FloatAttribute cy46d6Attr = projectionAttr.getChildByName( "cy46d6" );
            float cy46d6 = cy46d6Attr.getValue( 0.0f, false);
            FloatAttribute cx66d6Attr = projectionAttr.getChildByName( "cx66d6" );
            float cx66d6 = cx66d6Attr.getValue( 0.0f, false);
            FloatAttribute cy66d6Attr = projectionAttr.getChildByName( "cy66d6" );
            float cy66d6 = cy66d6Attr.getValue( 0.0f, false);

            ri->Projection( "bs_tdeLens", "int doDistort", &doDistort, "int model", &model,
                    "string distFile", &distFile, "float focal_length", &focal_length,
                    "float focus_distance", &focus_distance, "float filmBack_W", &filmBack_W,
                    "float filmBack_H", &filmBack_H, "float xOffset", &xOffset,
                    "float yOffset", &yOffset, "float distortion", &distortion,
                    "float anamorphicSqueeze", &anamorphicSqueeze, "float xCurvature", &xCurvature,
                    "float yCurvature", &yCurvature, "float quarticDist", &quarticDist,
                    "float dd2", &dd2, "float ud2", &ud2, "float vd2", &vd2, "float qd4", &qd4,
                    "float ud4", &ud4, "float vd4", &vd4, "float d6", &d6, "float d8", &d8,
                    "float phi", &phi, "float B", &B, "float cx02d2", &cx02d2,
                    "float cy02d2", &cy02d2, "float cx22d2", &cx22d2, "float cy22d2", &cy22d2,
                    "float cx04d4", &cx04d4, "float cy04d4", &cy04d4, "float cx24d4", &cx24d4,
                    "float cy24d4", &cy24d4, "float cx44d4", &cx44d4, "float cy44d4", &cy44d4,
                    "float rot", &rot, "float sqx", &sqx, "float sqy", &sqy,
                    "float cx06d6", &cx06d6, "float cy06d6", &cy06d6, "float cx26d6", &cx26d6,
                    "float cy26d6", &cy26d6, "float cx46d6", &cx46d6, "float cy46d6", &cy46d6,
                    "float cx66d6", &cx66d6, "float cy66d6", &cy66d6, RI_NULL );
            useDefaultProjection = false;
        }
        else if (projectionType == "OmnidirectionalStereo")
        {
            FloatAttribute interpupilaryDistanceAttr = projectionAttr.getChildByName( "interpupilaryDistance" );
            float interpupilaryDistance = interpupilaryDistanceAttr.getValue( 0.0f, false);
            ri->Projection("OmnidirectionalStereo", "float interpupilaryDistance", &interpupilaryDistance, RI_NULL);
            useDefaultProjection = false;
        }
    }

    // Emit another RiProjection "after" the RiProjection above for the
    // projection modifier.
    StringAttribute projectionModAttr = 
                projectionAttr.getChildByName("projectionModifier");
    if(projectionModAttr.isValid())
    {
        if (useDefaultProjection == true)
        {
            // If the projection type is default, we need to add the default
            // camera projection first before another RiProjection for the
            // projection modifier is emitted.
            addDefaultCameraProjection(cameraSettings);
            
            // Set this to false so that we won't call 
            // RiProjection in PRManRender::addCamera.
            useDefaultProjection = false;
        }

        std::string projectionModifier = projectionModAttr.getValue("none", false);
        if (projectionModifier == "PxrRollingShutter")
        {
            FloatAttribute angleAttr = projectionAttr.getChildByName("angle");
            float angle = angleAttr.getValue(90.f, false );
            ri->Projection((char*)"PxrRollingShutter", (char*)"float angle", 
                            &angle, RI_NULL);
        }
    }

    return useDefaultProjection;
}


int PRManRender::addCamera(FnKatRender::CameraSettings* cameraSettings,
                           PRManRenderSettings& renderSettings,
                           bool applyRenderSettings,
                           bool useDefaultProjection,
                           bool fovMotionBlur)
{
    if (!cameraSettings)
        return 1;

    Ri* ri = m_sharedState.GetRi();

    if (useDefaultProjection == true) 
    {
        addDefaultCameraProjection(cameraSettings);
    }
    FnScenegraphIterator rootIterator = getRootIterator();
    
    // add any projection plugins that are assigned to /root. These may be projetion
    // modifiers so they need to be added here (after the default camera) rather
    // than in addCameraProjection
    m_sharedState.WriteRI_ProjectionShaders(rootIterator);

    // ScreenWindow
    float screenWindow[4];
    cameraSettings->getScreenWindow(screenWindow);
    ri->ScreenWindow(screenWindow[0], screenWindow[2], 
                     screenWindow[1], screenWindow[3]);

    // Clipping
    float clipping[2];
    cameraSettings->getClipping(clipping);
    ri->Clipping(clipping[0], clipping[1]);

    // CropWindow
    if (applyRenderSettings)
        addCropWindow(renderSettings);

    // Camera
    ri->Scale(1.0, 1.0, -1.0);

    ri->ArchiveRecord((char*)"comment", (char*)"Camera: %s", 
                      cameraSettings->getName().c_str());

    // Camera FOV motion blur
    if( fovMotionBlur == true && cameraSettings->getProjection() == "perspective"
        && renderSettings.getMaxTimeSamples() > 1 )
    {
        FnScenegraphIterator cameraIterator = 
                    rootIterator.getByPath( cameraSettings->getName());
        if (cameraIterator.isValid())
        {
            DoubleAttribute fovAttr = cameraIterator.getAttribute( "geometry.fov");
            if (fovAttr.isValid())
            {
                addFOVMotionBlur(fovAttr, cameraSettings->getFov());
            }
        }
    }

    m_sharedState.WriteRI_Transforms(cameraSettings->getTransforms());

    return 0;
}


int PRManRender::addAdditionalCameras(PRManRenderSettings& renderSettings)
{
    Ri* ri = m_sharedState.GetRi();

    FnKatRender::RenderSettings::CameraMap 
                additionalCameras = renderSettings.getAdditionalCameras();
    FnKatRender::RenderSettings::CameraMap::const_iterator it;
    for (it = additionalCameras.begin(); it != additionalCameras.end(); ++it)
    {
        ri->TransformBegin();

        FnKatRender::CameraSettings* cameraSettings = (*it).second;
        addCamera(cameraSettings, renderSettings);

        std::string cameraName = cameraSettings->getName();
        ri->Camera(const_cast<RtToken>(cameraName.c_str()), RI_NULL);

        ri->TransformEnd();
    }

    return 0;
}

int PRManRender::addFrameCamera(FnScenegraphIterator sgi,
                                 PRManGlobalSettings& globalSettings,
                                 PRManRenderSettings& renderSettings)
{
    Ri* ri = m_sharedState.GetRi();

    const std::string cameraPath = globalSettings.getFrameBeginCameraPath();
    if (!cameraPath.empty())
    {
        // add DOF and Format if present
        FnScenegraphIterator cameraIterator = sgi.getByPath(cameraPath);
        if (cameraIterator.isValid())
        {
            FnAttribute::GroupAttribute cameraAttr =
                    cameraIterator.getAttribute(
                            "prmanGlobalStatements.camera");
            
            FnScenegraphIterator rootIterator = sgi.getRoot();
            
            if (!cameraAttr.isValid())
            {
                cameraAttr = cameraIterator.getAttribute(
                        "prmanGlobalStatements.camera");
            }
            
            if (cameraAttr.isValid())
            {
                m_sharedState.ApplyCameraGlobalSettings(cameraAttr);
            }
            
            FnAttribute::GroupAttribute renderSettingsAttr =
                    cameraIterator.getAttribute("renderSettings");
            
            if (!renderSettingsAttr.isValid())
            {
                renderSettingsAttr = rootIterator.getAttribute(
                        "renderSettings");
            }
            
            if (renderSettingsAttr.getChildByName("resolution").isValid())
            {
                PRManRenderSettings renderSettings(cameraIterator);
                addFormat( renderSettings );
            }
            else
            {
                PRManRenderSettings renderSettings(sgi);   
                addFormat( renderSettings );
            }
        }
        
        
        int overscan[4];
        renderSettings.getOverscan(overscan);

        int displayWindow[4];
        renderSettings.getDisplayWindow(displayWindow);

        FnKatRender::CameraSettings frameBeginCamera(sgi, cameraPath);
        frameBeginCamera.initialise(displayWindow, overscan, 0);

        ri->ArchiveRecord((char*)"comment", (char*)"Frame Begin Camera: %s", 
                          cameraPath.c_str());
        ri->Orientation(RI_RH);
        addCamera(&frameBeginCamera, renderSettings, false);
    }

    return 0;
}


void PRManRender::addLodrange(PRManGlobalSettings& globalSettings)
{
    Ri* ri = m_sharedState.GetRi();

    // prman has no notion of "default" lodrange, need to push down values
    // when rerendering whether they are enabled in the scene graph or not

    int def[2] = { 0, 4 };
    std::vector<int> lodrange(def, def + 2);

    FnAttribute::GroupAttribute globalSettingsAttr = 
                    globalSettings.getGlobalSettingsAttr();
    GroupAttribute rerenderAttr = globalSettingsAttr.getChildByName("rerender");
    if (rerenderAttr.isValid())
    {
        IntAttribute lodrangeAttr = rerenderAttr.getChildByName("lodrange");
        if (lodrangeAttr.isValid())
        {
            IntConstVector lodrangeOverride = lodrangeAttr.getNearestSample(0.f);
            if (lodrangeOverride.size() == 2)
            {
                lodrange[0] = lodrangeOverride[0];
                lodrange[1] = lodrangeOverride[1];
            }
        }
    }

    ri->Option((char*)"rerender", (char*)"int[2] lodrange", 
               (RtPointer)lodrange.data(), RI_NULL);
}

// Combine KatanaProcedural with ScenegraphLocationDelegate plugins

#include <PRManSphereScenegraphLocationDelegate.h>
#include <PRManPolymeshScenegraphLocationDelegate.h>
#include <PRManSubdmeshScenegraphLocationDelegate.h>
#include <PRManCurvesScenegraphLocationDelegate.h>
#include <PRManNurbspatchScenegraphLocationDelegate.h>
#include <PRManPointCloudScenegraphLocationDelegate.h>
#include <PRManSpheresScenegraphLocationDelegate.h>
#include <PRManBrickMapScenegraphLocationDelegate.h>
#include <PRManProceduralScenegraphLocationDelegate.h>
#include <PRManRibArchiveScenegraphLocationDelegate.h>
#include <PRManVolumeScenegraphLocationDelegate.h>

DEFINE_SCENEGRAPH_LOCATION_DELEGATE_PLUGIN(PRManSphereScenegraphLocationDelegate)
DEFINE_SCENEGRAPH_LOCATION_DELEGATE_PLUGIN(PRManPolymeshScenegraphLocationDelegate)
DEFINE_SCENEGRAPH_LOCATION_DELEGATE_PLUGIN(PRManSubdmeshScenegraphLocationDelegate)
DEFINE_SCENEGRAPH_LOCATION_DELEGATE_PLUGIN(PRManCurvesScenegraphLocationDelegate)
DEFINE_SCENEGRAPH_LOCATION_DELEGATE_PLUGIN(PRManNurbspatchScenegraphLocationDelegate)
DEFINE_SCENEGRAPH_LOCATION_DELEGATE_PLUGIN(PRManPointCloudScenegraphLocationDelegate)
DEFINE_SCENEGRAPH_LOCATION_DELEGATE_PLUGIN(PRManSpheresScenegraphLocationDelegate)
DEFINE_SCENEGRAPH_LOCATION_DELEGATE_PLUGIN(PRManBrickMapScenegraphLocationDelegate)
DEFINE_SCENEGRAPH_LOCATION_DELEGATE_PLUGIN(PRManProceduralScenegraphLocationDelegate)
DEFINE_SCENEGRAPH_LOCATION_DELEGATE_PLUGIN(PRManRibArchiveScenegraphLocationDelegate)
DEFINE_SCENEGRAPH_LOCATION_DELEGATE_PLUGIN(PRManVolumeScenegraphLocationDelegate)
DEFINE_RENDER_PLUGIN(PRManRender)

void registerPlugins() {
    REGISTER_PLUGIN(PRManPolymeshScenegraphLocationDelegate,    "PRMan21PolymeshScenegraphLocationDelegate",    0, 1);
    REGISTER_PLUGIN(PRManSubdmeshScenegraphLocationDelegate,    "PRMan21SubdmeshScenegraphLocationDelegate",    0, 1);
    REGISTER_PLUGIN(PRManSphereScenegraphLocationDelegate,      "PRMan21SphereScenegraphLocationDelegate",      0, 1);
    REGISTER_PLUGIN(PRManCurvesScenegraphLocationDelegate,      "PRMan21CurvesScenegraphLocationDelegate",      0, 1);
    REGISTER_PLUGIN(PRManNurbspatchScenegraphLocationDelegate,  "PRMan21NurbspatchScenegraphLocationDelegate",  0, 1);
    REGISTER_PLUGIN(PRManPointCloudScenegraphLocationDelegate,  "PRMan21PointCloudScenegraphLocationDelegate",  0, 1);
    REGISTER_PLUGIN(PRManSpheresScenegraphLocationDelegate,     "PRMan21SpheresScenegraphLocationDelegate",     0, 1);
    REGISTER_PLUGIN(PRManBrickMapScenegraphLocationDelegate,    "PRMan21BrickMapScenegraphLocationDelegate",    0, 1);
    REGISTER_PLUGIN(PRManProceduralScenegraphLocationDelegate,  "PRMan21ProceduralScenegraphLocationDelegate",  0, 1);
    REGISTER_PLUGIN(PRManRibArchiveScenegraphLocationDelegate,  "PRMan21RibArchiveScenegraphLocationDelegate",  0, 1);
    REGISTER_PLUGIN(PRManVolumeScenegraphLocationDelegate,      "PRMan21VolumeScenegraphLocationDelegate",      0, 1);
    REGISTER_PLUGIN(PRManRender, "prman", 0, 1);
}
